import os
from datetime import datetime

from pydoctordroid._event_exporter import HttpEventExporter
from pydoctordroid._events import create_event
from pydoctordroid._logger import setup_logger

_DRDROID_AUTH_TOKEN_ENV = 'DRDROID_AUTH_TOKEN'
_DRDROID_HOSTNAME_ENV = 'DRDROID_HOSTNAME'
_DRDROID_DEBUG = 'DRDROID_DEBUG'

logger = setup_logger(debug=os.environ.get(_DRDROID_DEBUG, False))


class DrDroid:
    def __init__(
            self,
            token: str = '',
            endpoint: str = 'https://ingest.drdroid.io',
            debug: bool = False,
            logger=None
    ):
        self._setup = False
        self._token = os.environ.get(_DRDROID_AUTH_TOKEN_ENV, token)
        self._hostname = os.environ.get(_DRDROID_HOSTNAME_ENV, endpoint)
        if logger:
            self._logger = logger
        elif os.environ.get(_DRDROID_DEBUG, debug):
            self._logger = setup_logger('drdroid.sdk.instance', True)
        else:
            self._logger = setup_logger('drdroid.sdk.instance', False)

        self._event_exporter = HttpEventExporter(self._token, self._hostname, self._logger)
        if self._event_exporter:
            self._setup = True
        return

    def publish(self, workflow: str, state: str, payload: dict = None, event_time: datetime = None):
        if not self._setup:
            return
        try:
            event = create_event(workflow, state, payload, event_time)
            self._event_exporter.export(event)
        except Exception as e:
            self._logger.debug(msg=f'Error exporting event: {e}')
        return


class _Global:
    def __init__(self):
        self._instance = DrDroid(logger=logger)

    def setup(self, *args, **kwargs):
        self._instance = DrDroid(*args, *kwargs)

    def publish(self, *args, **kwargs):
        self._instance.publish(*args, **kwargs)


drdroid = _Global()


def setup(token: str, endpoint: str):
    drdroid.setup(token, endpoint)


def publish(workflow: str, state: str, payload: dict = None, event_time: datetime = None):
    drdroid.publish(workflow, state, payload, event_time)



import logging
import sys


def setup_logger(name: str = 'drdroid.sdk.global', debug: bool = False) -> logging.Logger:
    logger = logging.getLogger(name)

    log_level = logging.DEBUG if debug else logging.INFO
    logger.setLevel(log_level)

    formatter = logging.Formatter(fmt='[drdroid] %(asctime)s %(levelname)s %(message)s')
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    return logger



import abc
from abc import abstractmethod

import urllib3

from pydoctordroid._events import serialize_events


class EventExporter(abc.ABC):
    @abstractmethod
    def export(self, event):
        pass


class HttpEventExporter(EventExporter):
    def __init__(self, token, hostname, logger):
        self._pool = urllib3.PoolManager(
            headers={
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {token}'
            },
            timeout=urllib3.Timeout(connect=1.0, read=2.0)
        )

        self._endpoint = f"{hostname}/w/agent/push_events"
        self._logger = logger

        return

    def export(self, event):
        payload = serialize_events([event])
        try:
            response = self._pool.request(
                "POST",
                self._endpoint,
                body=payload,
            )
            if response.status == 200:
                self._logger.debug(msg='Event exported successfully')
            elif response.status == 401:
                self._logger.debug(msg='Error exporting event: Request Unauthorized. Validate auth token.')

            response.close()
        except Exception as e:
            self._logger.debug(msg=f'{e}')



from collections.abc import Iterable
from datetime import datetime
from typing import Dict

_VALID_TYPES = [bool, str, int, float, datetime, list, tuple, set, dict, Iterable]


class ValueFactory:
    @staticmethod
    def value(obj):
        if obj is None:
            return None
        if isinstance(obj, bool):
            return BoolValue(obj)
        elif isinstance(obj, str):
            return StringValue(obj)
        elif isinstance(obj, int):
            return LongValue(obj)
        elif isinstance(obj, float):
            return DoubleValue(obj)
        elif isinstance(obj, bytes):
            return BytesValue(obj)
        elif isinstance(obj, datetime):
            return DatetimeValue(obj)
        elif isinstance(obj, dict):
            return DictValue(obj)
        elif isinstance(obj, Iterable):
            return ArrayValue(obj)
        return ObjValue(obj)


class InvalidValueTypeError(ValueError):
    pass


class Value:
    key = ''
    typecheck = None
    _value = None

    def __init__(self, value):
        if self.typecheck and not isinstance(value, self.typecheck):
            raise InvalidValueTypeError(f'expected value of type {self.typecheck}')
        self._value = self.process(value)

    @classmethod
    def process(cls, value):
        return value

    def value(self) -> Dict:
        return {
            self.key: self._value
        }


class StringValue(Value):
    key = 'string_value'
    typecheck = str


class BoolValue(Value):
    key = 'bool_value'
    typecheck = bool


class LongValue(Value):
    key = 'int_value'
    typecheck = int


class DoubleValue(Value):
    key = 'double_value'
    typecheck = float


class BytesValue(Value):
    key = 'bytes_value'
    typecheck = bytes


class KeyValue(Value):
    typecheck = Value

    def __init__(self, key: str, value: Value):
        super().__init__(value)
        self._key = key

    def value(self) -> Dict:
        return {
            'key': self._key,
            'value': self._value
        }


class DatetimeValue(StringValue):
    typecheck = datetime


class ArrayValue(Value):
    key = 'array_value'
    typecheck = Iterable

    @classmethod
    def process(cls, value):
        return [ValueFactory.value(elem) for elem in value if elem]

    def value(self) -> Dict:
        return {
            self.key: {
                'values': self._value
            }
        }


class KVListValue(Value):
    key = 'kvlist_value'

    def value(self) -> Dict:
        return {
            self.key: {
                'values': self._value
            }
        }


class DictValue(KVListValue):
    typecheck = dict

    @classmethod
    def process(cls, value):
        return [KeyValue(k, ValueFactory.value(v)) for k, v in value.items() if v and
                type(v) in _VALID_TYPES or isinstance(type, Iterable)]


class ObjValue(KVListValue):
    typecheck = None

    @classmethod
    def process(cls, value):
        return [KeyValue(k, ValueFactory.value(v)) for k, v in value.__dict__.items() if v and
                type(v) in _VALID_TYPES or isinstance(type, Iterable)]


def process_payload(payload: Dict):
    return [KeyValue(key, ValueFactory.value(value)) for key, value in payload.items() if
            isinstance(key, str) and value]



from setuptools import find_packages, setup

# with open("README.md", "r", encoding = "utf-8") as fh:
#     long_description = fh.read()

setup(
    name='pydoctordroid',
    setup_requires=['wheel'],
    packages=find_packages(),
    include_package_data=True,
    version='0.1.1',
    description='Code Marker Library from Doctor Droid to generate custom events',
    author='Dipesh Mittal',
    author_email="dipesh@drdroid.io",
    license='MIT',
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["urllib3>=1.26.11"],
)



import json
from datetime import datetime, timezone
from typing import Any, Dict, List

from pydoctordroid._value import process_payload, Value

EventType = Dict


def create_event(workflow: str, state: str, payload: dict, event_time: datetime = None) -> EventType:
    return {
        'workflow': {'name': workflow},
        'state': state,
        'timestamp': (event_time or datetime.now(timezone.utc)),
        'kvs': process_payload(payload or {})
    }


class EventEncoder(json.JSONEncoder):
    fmt = '%Y-%m-%dT%H:%M:%S.%fZ'

    def default(self, o: Any) -> Any:
        if isinstance(o, Value):
            return o.value()
        elif isinstance(o, datetime):
            return o.astimezone(timezone.utc).strftime(self.fmt)
        elif isinstance(o, set):
            return list(o)

        return super().default(o)


def serialize_events(events: List[EventType]):
    event_data = {'data': {'events': events}}
    return json.dumps(event_data, separators=(',', ':'), cls=EventEncoder)
